// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.CommandLine
open System.CommandLine.Invocation
open System.IO
open System.Threading.Tasks

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let writeTempCsproj (directory: string) (dotnetVersionString: string) =
    let csprojPath = Path.Combine(directory, $"Project_{dotnetVersionString}.csproj")

    let csprojContent =
        $"""<Project Sdk="Microsoft.NET.Sdk">
  <PropertyGroup>
    <NoWarn>\$(NoWarn);NETSDK1138</NoWarn>
    <RollForward>Major</RollForward>
    <TargetFramework>net{dotnetVersionString}</TargetFramework>
  </PropertyGroup>
</Project>"""

    use writer = new StreamWriter(csprojPath)

    do writer.Write(csprojContent)
    writer.Flush()

let getSdkVersionChain (min: double) (max: double) =
    let generated: List<float> = [ System.Math.Ceiling(min) .. max ]

    float min :: float max :: generated |> List.distinct |> List.sort

// TODO: "gdmt-restore" never fails, use heuristic to determine if it has failed?
let runGdmtRestore (args: List<string>) =
    ExecProcess("gdmt-restore" :: args).Run().Check()

let commandName = "gdmt-checkcore"
let commandDescription = "check versions of NuGet pkgs bound to SDK version"

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(commandName)

    rootCommand.Name <- commandName
    rootCommand.Description <- commandDescription

    let sdkMinVersionOption =
        new Option<string>([| "-m"; "--sdk-min" |], "minimal SDK version")

    sdkMinVersionOption.SetDefaultValue "6.0"

    let sdkMaxVersionOption =
        new Option<string>([| "-e"; "--sdk-ver" |], "targeted SDK version")

    sdkMaxVersionOption.SetDefaultValue "8.0"

    let sdkExecutableOption =
        new Option<string>([| "-x"; "--sdk-exe" |], "SDK executable path")

    let rootCommandHandler =
        Func<InvocationContext, Task>(fun context ->
            let operation =
                async {
                    let options = context.ParseResult

                    let sdkMinVersion = double (options.GetValueForOption sdkMinVersionOption)
                    let sdkMaxVersion = double (options.GetValueForOption sdkMaxVersionOption)

                    let tempGuid = Guid.NewGuid().ToString()
                    let tempName = $"gdmt_check_core_{tempGuid}"
                    let tempPath = Path.Combine(Path.GetTempPath(), tempName)

                    SimpleLog.LogMessage SimpleLog.Debug $"Temp path is: {tempPath}"

                    for sdkVersion in getSdkVersionChain sdkMinVersion sdkMaxVersion do
                        let sdkVersionString = sprintf "%.1f" sdkVersion
                        let sdkVersionTempPath = Path.Combine(tempPath, sdkVersionString)

                        SimpleLog.LogMessage SimpleLog.Debug $"Starting restore for .NET {sdkVersionString}"

                        Directory.CreateDirectory sdkVersionTempPath |> ignore
                        writeTempCsproj sdkVersionTempPath sdkVersionString

                        let mutable gdmtRestoreArgs = [ "-p"; sdkVersionTempPath ]

                        match options.GetValueForOption sdkExecutableOption with
                        | emptyValue when String.IsNullOrEmpty emptyValue -> ()
                        | nonEmptyValue ->
                            gdmtRestoreArgs <- List.append gdmtRestoreArgs [ "--sdk-exe"; nonEmptyValue ]

                        runGdmtRestore gdmtRestoreArgs
                        Directory.Delete(sdkVersionTempPath, true)

                    Directory.Delete(tempPath, true)

                    ()
                }

            Async.StartAsTask(operation, TaskCreationOptions.None))

    rootCommand.AddOption sdkMinVersionOption
    rootCommand.AddOption sdkMaxVersionOption
    rootCommand.AddOption sdkExecutableOption

    rootCommand.SetHandler rootCommandHandler

    rootCommand.Invoke(argv)
