// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Text.RegularExpressions
open System.Threading.Tasks

open LibGit2Sharp
open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let getChangedFiles (repoTopDirectory: string) =
    use repo = new Repository(repoTopDirectory)

    repo.Diff.Compare<TreeChanges>(repo.Head.Tip.Tree, DiffTargets.WorkingDirectory)
    |> Seq.map (fun change -> Path.GetFileName change.Path)
    |> List.ofSeq

let updateEbuildManifests () =
    SimpleLog.LogMessage SimpleLog.Debug "Updating manifests"
    ExecProcess([ "pkgdev"; "manifest" ]).Run().Check()
    SimpleLog.LogMessage SimpleLog.Success "Successfully updated manifests"

let gitAddAll () =
    ExecProcess([ "git"; "add"; "--all"; "." ]).Run().Check()

let pkgcheckScan () =
    SimpleLog.LogMessage SimpleLog.Debug "Scanning with pkgcheck"
    ExecProcess([ "pkgcheck"; "scan"; "--exit"; "error" ]).Run().Check()
    SimpleLog.LogMessage SimpleLog.Success "Scan with pkgcheck succeeded"

let isEbuild (path: string) = Regex.IsMatch(path, ".*\.ebuild$")

let isLiveEbuild (path: string) =
    isEbuild path && Regex.IsMatch(path, ".*9999\..*")

let testEbuilds (ebuildFiles: seq<string>) =
    for ebuild in ebuildFiles do
        match ebuild with
        | f when File.Exists f ->
            SimpleLog.LogMessage SimpleLog.Debug $"Testing ebuild: {f}"

            ExecProcess([ "gdmt-ebuild"; f; "clean"; "install"; "clean" ]).Run().Check()

            SimpleLog.LogMessage SimpleLog.Success $"Ebuild test passed for: {f}"
        | f -> SimpleLog.LogMessage SimpleLog.Debug $"Ebuild file does not exist: {f}"

let guardCommitMessage (commitMessage: string) =
    if String.IsNullOrEmpty commitMessage then
        "the effective commit message is empty" |> Exception |> raise

let createTags (tagType: string) (sources: string[]) =
    sources
    |> Array.map (fun source ->
        match source with
        | s when String.IsNullOrEmpty s -> ""
        | s when Regex.IsMatch(s, "http.*") -> $"{tagType}: {s}\n"
        | s -> $"{tagType}: https://bugs.gentoo.org/{s}\n")
    |> String.concat ""

let commandName = "gdmt-commit"
let commandDescription = "commit ebuild packages"

// Git / commit message

let dryRunOption = new Option<bool>([| "-d"; "--dry" |], "do not make commits")
dryRunOption.SetDefaultValue false

let gitCommitMessageOption =
    new Option<string>([| "-m"; "--message" |], "commit message")

let gitCommitMessageAuxOption =
    new Option<string>([| "-M"; "--message-aux" |], "auxiliary commit message")

let gitSignOption = new Option<bool>([| "-S"; "--sign" |], "sign the create commit")

gitSignOption.SetDefaultValue false

// Tests

let scanOption = new Option<bool>([| "-s"; "--scan" |], "scan using pkgcheck")
scanOption.SetDefaultValue false

let testOption = new Option<bool>([| "-t"; "--test" |], "test using ebuild")
testOption.SetDefaultValue false

// Bump

let bumpOption =
    new Option<string>([| "-B"; "--bump" |], "set message to \"bump to VERSION\"")

let bumpCopyOption =
    new Option<bool>([| "-C"; "--copy" |], "if doing a bump copy latest ebuild")

bumpCopyOption.SetDefaultValue false

// Drop

let dropOption =
    new Option<string>([| "-D"; "--drop" |], "set message to \"drop old VERSION\"")

let dropRemoveOption =
    new Option<bool>([| "-R"; "--remove" |], "if doing a drop, remove ebuild")

dropRemoveOption.SetDefaultValue false

// Bugs

let bugOption =
    new Option<string>([| "-b"; "--bug" |], "reference a Gentoo BugZilla or upstream bug/URL")

let closesOption =
    new Option<string>([| "-c"; "--closes" |], "closes a Gentoo BugZilla bug")

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(commandName)

    rootCommand.Name <- commandName
    rootCommand.Description <- commandDescription

    let rootCommandHandler =
        Func<InvocationContext, Task>(fun context ->
            let operation =
                async {
                    let options = context.ParseResult

                    let dryRun = options.GetValueForOption dryRunOption

                    let gitCommitMessage = options.GetValueForOption gitCommitMessageOption
                    let gitCommitMessageAux = options.GetValueForOption gitCommitMessageAuxOption
                    let gitSign = options.GetValueForOption gitSignOption

                    let scan = options.GetValueForOption scanOption
                    let test = options.GetValueForOption testOption

                    let bump = options.GetValueForOption bumpOption
                    let bumpCopy = options.GetValueForOption bumpCopyOption
                    let drop = options.GetValueForOption dropOption
                    let dropRemove = options.GetValueForOption dropRemoveOption

                    let bugSources = options.GetValueForOption bugOption
                    let closesSources = options.GetValueForOption closesOption

                    let mutable commitMessage = ""

                    let packageFullName =
                        let segments = ("." |> Path.GetFullPath).Split('/') |> Array.rev |> Array.toList

                        match segments with
                        | segment1 :: segment2 :: _ -> Path.Combine(segment2, segment1)
                        | _ -> "wrong list length" |> Exception |> raise

                    let packageName = Path.GetFileName packageFullName

                    if not (String.IsNullOrEmpty bump) then
                        commitMessage <- $"{packageFullName}: bump to {bump}"

                        if bumpCopy then
                            let latestEbuilds =
                                Directory.GetFiles(".", "*.ebuild")
                                |> Array.filter (isLiveEbuild >> not)
                                |> Array.sort
                                |> Array.rev

                            match latestEbuilds with
                            | [||] -> "no latest ebuild available" |> Exception |> raise
                            | _ -> ()

                            let latest = latestEbuilds |> Array.head
                            let updated = $"{packageName}-{bump}.ebuild"

                            SimpleLog.LogMessage SimpleLog.Debug $"Copying {latest} into {updated}"
                            File.Copy(latest, updated)

                            if not dryRun then
                                ExecProcess([ "git"; "add"; updated ]).Run().Check()

                    if not (String.IsNullOrEmpty drop) then
                        commitMessage <- $"{packageFullName}: drop old {drop}"

                        if dropRemove then
                            let old = $"{packageName}-{drop}.ebuild"

                            SimpleLog.LogMessage SimpleLog.Debug $"Removing {old}"
                            File.Delete(old)

                    let changes = getChangedFiles "../.."
                    let changedEbuilds = changes |> List.filter isEbuild

                    for change in changes do
                        SimpleLog.LogMessage SimpleLog.Debug $"Detected change: {change}"

                    for change in changedEbuilds do
                        SimpleLog.LogMessage SimpleLog.Debug $"Modified ebuild: {change}"

                    updateEbuildManifests ()

                    if not dryRun then
                        gitAddAll ()

                    if scan then
                        pkgcheckScan ()

                    if not (String.IsNullOrEmpty gitCommitMessage) then
                        commitMessage <- $"{packageFullName}: {gitCommitMessage}"

                    guardCommitMessage commitMessage

                    if test then
                        match changedEbuilds with
                        | [] -> SimpleLog.LogMessage SimpleLog.Debug "No changed ebuilds, nothing to do"
                        | _ -> testEbuilds changedEbuilds

                    let bugTags =
                        match bugSources with
                        | s when String.IsNullOrEmpty s -> ""
                        | s -> createTags "Bug" (s.Split [| ' ' |])

                    let closesTags =
                        match closesSources with
                        | s when String.IsNullOrEmpty s -> ""
                        | s -> createTags "Closes" (s.Split [| ' ' |])

                    let commitMessageFull =
                        commitMessage
                        + "\n"
                        + match gitCommitMessageAux with
                          | s when String.IsNullOrEmpty s -> ""
                          | s -> "\n" + s + "\n"
                        + "\n"
                        + bugTags
                        + closesTags

                    if not dryRun then
                        let gitCommitArgs =
                            [ "git"; "commit"; "-m"; $"'{commitMessageFull}'" ]
                            @ (if gitSign then [ "--signoff" ] else [])

                        ExecProcess(gitCommitArgs).Run().Check()

                    ()
                }

            Async.StartAsTask(operation, TaskCreationOptions.None))

    rootCommand.AddOption dryRunOption

    rootCommand.AddOption gitCommitMessageOption
    rootCommand.AddOption gitCommitMessageAuxOption
    rootCommand.AddOption gitSignOption

    rootCommand.AddOption scanOption
    rootCommand.AddOption testOption

    rootCommand.AddOption bumpOption
    rootCommand.AddOption bumpCopyOption
    rootCommand.AddOption dropOption
    rootCommand.AddOption dropRemoveOption

    rootCommand.AddOption bugOption
    rootCommand.AddOption closesOption

    rootCommand.SetHandler rootCommandHandler

    rootCommand.Invoke(argv)
