// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.IO
open System.Threading.Tasks

open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let commandName = "gdmt-ebuild"
let commandDescription = "wrapper to run the ebuild command faster and for multiple files"

// Ebuild command

let ebuildPathArgument = new Argument<string>("ebuild-path", "ebuild file path or directory of ebuilds")

let ebuildCommandArgument =
    new Argument<string[]>("ebuild-command", "ebuild command(s), e.g. prepare, compile")

ebuildCommandArgument.SetDefaultValue [| "manifest"; "clean"; "install" |]

// Package manager (Portage)

let portageFeaturesOption =
    new Option<string>([| "-f"; "--features" |], "Portage features")

portageFeaturesOption.SetDefaultValue "test"

let portageUseOption = new Option<string>([| "-u"; "--use" |], "Portage USE flags")
portageUseOption.SetDefaultValue "test"

let forceTestsOption = new Option<bool>([| "-t"; "--test" |], "force running tests")
forceTestsOption.SetDefaultValue false

// System stress

let nicenessOption = new Option<int>([| "-n"; "--nice" |], "niceness level")
nicenessOption.SetDefaultValue 19

let nproc = Environment.ProcessorCount

let jobsOption = new Option<int>([| "-j"; "--jobs" |], "max number of jobs")
jobsOption.SetDefaultValue nproc

let loadAverageOption =
    new Option<int>([| "-l"; "--load" |], "max number of load average")

loadAverageOption.SetDefaultValue nproc

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(commandName)

    rootCommand.Name <- commandName
    rootCommand.Description <- commandDescription

    let rootCommandHandler =
        Func<InvocationContext, Task>(fun context ->
            let operation =
                async {
                    let options = context.ParseResult

                    let ebuildPath = options.GetValueForArgument(ebuildPathArgument)

                    let ebuildCommand =
                        options.GetValueForArgument(ebuildCommandArgument) |> Array.toList

                    let portageFeatures = options.GetValueForOption(portageFeaturesOption)
                    let portageUse = options.GetValueForOption(portageUseOption)
                    let forceTests = options.GetValueForOption(forceTestsOption)

                    let niceness = options.GetValueForOption(nicenessOption)
                    let jobs = options.GetValueForOption(jobsOption)
                    let loadAverage = options.GetValueForOption(loadAverageOption)

                    let ebuilds =
                        match ebuildPath with
                        | p when Directory.Exists p -> Directory.GetFiles(ebuildPath, "*.ebuild")
                        | p when File.Exists p -> [| p |]
                        | _ -> $"The ebuild file does not exist, given {ebuildPath}" |> Exception |> raise

                    let portageTestFlag = if forceTests then " test" else ""

                    let envMap =
                        [ ("ALLOW_TEST", (if forceTests then "all" else ""))
                          ("FEATURES", portageFeatures + portageTestFlag)
                          ("MAKEOPTS", $"-j {jobs} -l {loadAverage}")
                          ("USE", portageUse + portageTestFlag) ]

                    for ebuild in ebuilds do
                        let args = [ "nice"; "-n"; $"{niceness}"; "ebuild"; ebuild ] @ ebuildCommand

                        SimpleLog.LogMessage SimpleLog.Debug $"Running ebuild command for {ebuild}"
                        ExecProcess(args, envMap).Run().Check()
                        SimpleLog.LogMessage SimpleLog.Success $"Ebuild command for {ebuild} finished successfully"

                    ()
                }

            Async.StartAsTask(operation, TaskCreationOptions.None))

    rootCommand.AddArgument ebuildPathArgument
    rootCommand.AddArgument ebuildCommandArgument

    rootCommand.AddOption portageFeaturesOption
    rootCommand.AddOption portageUseOption
    rootCommand.AddOption forceTestsOption

    rootCommand.AddOption nicenessOption
    rootCommand.AddOption jobsOption
    rootCommand.AddOption loadAverageOption

    rootCommand.SetHandler rootCommandHandler

    rootCommand.Invoke(argv)
