// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.CommandLine
open System.CommandLine.Invocation
open System.IO
open System.Threading.Tasks

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let commandName = "gdmt-restore"
let commandDescription = "maintenance tool to restore projects"

let linuxRuntimes =
    [
      // generic
      "linux"

      // glibc
      "linux-arm"
      "linux-arm64"
      "linux-x64"
      "linux-x86"

      // musl
      "linux-musl-arm"
      "linux-musl-arm64"
      "linux-musl-x64"
      "linux-musl-x86" ]

let dotnetRuntimeNugets =
    [ "microsoft.aspnetcore.app.ref"
      "microsoft.aspnetcore.app.runtime."
      "microsoft.netcore.app.host."
      "microsoft.netcore.app.ref"
      "microsoft.netcore.app.runtime." ]

let findSdkExecutable (sdkVersion: string) : string =
    let srcExe = $"/usr/bin/dotnet-{sdkVersion}"
    let binExe = $"/usr/bin/dotnet-bin-{sdkVersion}"

    if File.Exists(srcExe) then
        srcExe
    elif File.Exists(binExe) then
        binExe
    else
        "no suitable dotnet sdk executable was found" |> Exception |> raise

let getFileDirectory (path: string) =
    match path with
    | d when Directory.Exists d -> d
    | f when File.Exists f -> Path.GetDirectoryName(f)
    | _ -> $"path {path} does not exist" |> Exception |> raise

let restoreForRuntime
    (sdkExe: string)
    (runtime: string)
    (projectPath: string)
    (tmpDir: string)
    (nugetPackagesDir: string)
    (extraRestoreArgs: string)
    =

    let env =
        [ ("PWD", getFileDirectory projectPath)
          ("TMPDIR", tmpDir)
          ("NUGET_PACKAGES", nugetPackagesDir)

          ("DOTNET_CLI_TELEMETRY_OPTOUT", "1")
          ("DOTNET_NOLOGO", "1")
          ("DOTNET_SKIP_FIRST_TIME_EXPERIENCE", "1")
          ("MSBUILDDISABLENODEREUSE", "1")
          ("UseSharedCompilation", "false") ]

    let args =
        [ sdkExe
          "restore"
          "--force-evaluate"
          "--verbosity"
          "quiet"
          "--runtime"
          runtime
          extraRestoreArgs
          projectPath ]

    let execProcess = ExecProcess(args, env)

    execProcess.Run().Success

let forEachRuntime (lambdaForRuntime: string -> bool) : unit =
    linuxRuntimes
    |> List.map (fun runtime ->
        SimpleLog.LogMessage SimpleLog.Debug $"Performing restore for runtime: {runtime}"

        let exitSuccess = lambdaForRuntime runtime

        (runtime, exitSuccess))
    |> List.map (fun (runtime, exitSuccess) ->
        match exitSuccess with
        | true -> SimpleLog.LogMessage SimpleLog.Success $"Restore for runtime {runtime} passed successfully"
        | _ -> SimpleLog.LogMessage SimpleLog.Error $"Restore for runtime {runtime} has failed")
    |> ignore

let gatherNugets (nugetPackagesDir: string) : List<string> =
    try
        Directory.GetDirectories(nugetPackagesDir, "*", SearchOption.AllDirectories)
        |> Seq.map (fun path -> path.Substring(nugetPackagesDir.Length + 1))
        |> Seq.map (fun path -> path.Split('/'))
        |> Seq.filter (fun pathElements -> pathElements.Length = 2)
        |> Seq.map (fun pathElements -> $"{pathElements.[0]}@{pathElements.[1]}")
        |> Seq.filter (fun atNuget -> not (dotnetRuntimeNugets |> List.exists atNuget.Contains))
        |> Seq.sort
        |> Seq.toList
    with ex ->
        SimpleLog.LogMessage SimpleLog.Error ex.Message
        []

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(commandName)

    rootCommand.Name <- commandName
    rootCommand.Description <- commandDescription

    let projectPathOption = new Option<string>([| "-p"; "--project" |], "project path")

    projectPathOption.SetDefaultValue "."

    let sdkVersionOption = new Option<string>([| "-e"; "--sdk-ver" |], "SDK version")

    sdkVersionOption.SetDefaultValue "8.0"

    let sdkExecutableOption =
        new Option<string>([| "-x"; "--sdk-exe" |], "SDK executable path")

    let cacheDirOption =
        new Option<string>([| "-c"; "--cache" |], "cache directory to use")

    let extraRestoreArgsOption =
        new Option<string>([| "-r"; "--extra-r" |], "extra restore arguments")

    let rootCommandHandler =
        Func<InvocationContext, Task>(fun context ->
            let operation =
                async {
                    let options = context.ParseResult

                    let projectPath = options.GetValueForOption(projectPathOption) |> Path.GetFullPath

                    if not (File.Exists(projectPath) || Directory.Exists(projectPath)) then
                        $"project {projectPath} does not exist" |> Exception |> raise

                    let cacheDir =
                        match options.GetValueForOption(cacheDirOption) with
                        | emptyValue when String.IsNullOrEmpty emptyValue ->
                            Path.Combine(projectPath |> getFileDirectory, ".cache")
                        | nonEmptyValue -> nonEmptyValue |> Path.GetFullPath

                    let tmpDir = Path.Combine(cacheDir, "tmpdir")
                    let nugetPackagesDir = Path.Combine(cacheDir, "nuget_packages")

                    let sdkExecutable =
                        match options.GetValueForOption(sdkExecutableOption) with
                        | emptyValue when String.IsNullOrEmpty emptyValue ->
                            sdkVersionOption |> options.GetValueForOption |> findSdkExecutable
                        | nonEmptyValue -> nonEmptyValue

                    let extraRestoreArgs = extraRestoreArgsOption |> options.GetValueForOption

                    Directory.CreateDirectory tmpDir |> ignore

                    SimpleLog.LogMessage SimpleLog.Debug $"SDK executable path:       {sdkExecutable}"
                    SimpleLog.LogMessage SimpleLog.Debug $"Project path:              {projectPath}"
                    SimpleLog.LogMessage SimpleLog.Debug $"Cache directory:           {cacheDir}"
                    SimpleLog.LogMessage SimpleLog.Debug $"TMPDIR directory:          {tmpDir}"
                    SimpleLog.LogMessage SimpleLog.Debug $"NUGET_PACKAGES directory:  {nugetPackagesDir}"

                    forEachRuntime (fun runtime ->
                        restoreForRuntime sdkExecutable runtime projectPath tmpDir nugetPackagesDir extraRestoreArgs)

                    printfn "Those are the gathered NuGet packages:"
                    printfn ""

                    gatherNugets nugetPackagesDir
                    |> List.map (fun nuget -> printfn "%s" nuget)
                    |> ignore

                    ()
                }

            Async.StartAsTask(operation, TaskCreationOptions.None))

    rootCommand.AddOption projectPathOption
    rootCommand.AddOption sdkVersionOption
    rootCommand.AddOption sdkExecutableOption
    rootCommand.AddOption cacheDirOption
    rootCommand.AddOption extraRestoreArgsOption

    rootCommand.SetHandler rootCommandHandler

    rootCommand.Invoke(argv)
