// Copyright 2023-2024 Gentoo Authors
// Distributed under the terms of the GNU General Public License v2

open System
open System.Threading.Tasks

open System.CommandLine
open System.CommandLine.Invocation

open Gdmt.Shared.ExecProcess
open Gdmt.Shared.SimpleLog

let commandName = "gdmt-sync"
let commandDescription = "synchronize the ebuild repository"

[<EntryPoint>]
let main argv =
    let rootCommand = RootCommand(commandName)

    rootCommand.Name <- commandName
    rootCommand.Description <- commandDescription

    // Remotes

    let remotesArgument = new Argument<string[]>("remotes", "remotes to push to")

    remotesArgument.SetDefaultValue [| "origin" |]

    // Git

    let gitPullFromRemoteOption =
        Option<string>([| "-f"; "--from" |], "pull from specified remote")

    gitPullFromRemoteOption.SetDefaultValue "origin"

    let gitSignOrigin =
        Option<bool>([| "-s"; "--sign" |], "sign when pushing to origin remote")

    gitSignOrigin.SetDefaultValue false

    // System stress

    let nproc = Environment.ProcessorCount

    let jobsOption = new Option<int>([| "-j"; "--jobs" |], "max number of jobs")

    jobsOption.SetDefaultValue nproc

    let loadAverageOption =
        new Option<int>([| "-l"; "--load" |], "max number of load average")

    loadAverageOption.SetDefaultValue nproc

    let rootCommandHandler =
        Func<InvocationContext, Task>(fun context ->
            let operation =
                async {
                    let options = context.ParseResult

                    let remotes = options.GetValueForArgument remotesArgument

                    let gitPullFromRemote = options.GetValueForOption gitPullFromRemoteOption
                    let gitSignOrigin = options.GetValueForOption gitSignOrigin

                    let jobs = options.GetValueForOption jobsOption
                    let loadAverage = options.GetValueForOption loadAverageOption

                    SimpleLog.LogMessage SimpleLog.Debug $"Scanning with pkgcheck"
                    ExecProcess([ "pkgcheck"; "scan"; "--commits"; "--exit=error" ]).Run().Check()
                    SimpleLog.LogMessage SimpleLog.Success "Scan succeeded"

                    SimpleLog.LogMessage SimpleLog.Debug $"Pulling from remote {gitPullFromRemote}"
                    ExecProcess([ "git"; "pull"; gitPullFromRemote ]).Run().Check()
                    SimpleLog.LogMessage SimpleLog.Success $"Pull from {gitPullFromRemote} succeeded"

                    if gitSignOrigin && Array.contains "origin" remotes then
                        SimpleLog.LogMessage SimpleLog.Debug "Pushing to remote: origin"
                        ExecProcess([ "git"; "push"; "origin"; "--sign" ]).Run().Check()
                        SimpleLog.LogMessage SimpleLog.Success "Push to origin succeeded"

                    for remote in remotes do
                        if gitSignOrigin && remote = "origin" then
                            ()
                        else
                            SimpleLog.LogMessage SimpleLog.Debug $"Pushing to remote: {remote}"
                            ExecProcess([ "git"; "push"; remote ]).Run().Check()
                            SimpleLog.LogMessage SimpleLog.Success $"Push to {remote} succeeded"

                    let egencacheArgs =
                        [ "egencache"
                          "--jobs"
                          $"{jobs}"
                          "--load-average"
                          $"{loadAverage}"
                          "--repo"
                          "gentoo"
                          "--update"
                          "--update-pkg-desc-index"
                          "--update-use-local-desc" ]

                    SimpleLog.LogMessage SimpleLog.Debug "Updating ebuild cache"
                    ExecProcess(egencacheArgs).Run().Check()
                    SimpleLog.LogMessage SimpleLog.Success "Cache update succeeded"

                    SimpleLog.LogMessage SimpleLog.Debug "Updating eix database"
                    ExecProcess([ "eix-update"; "--quiet" ]).Run().Check()
                    SimpleLog.LogMessage SimpleLog.Success "Database update succeeded"

                    SimpleLog.LogMessage SimpleLog.Debug "Refreshing the git repository index"
                    ExecProcess([ "git"; "update-index"; "--refresh" ]).Run().Check()
                    SimpleLog.LogMessage SimpleLog.Success "Refresh succeeded"

                    SimpleLog.LogMessage SimpleLog.Debug "Resolving updates"
                    ExecProcess([ "eselect"; "news"; "list"; "new" ]).Run().Check()

                    let eixUpdateArgs = [ "eix"; "--upgrade"; "--only-names" ]
                    ExecProcess(eixUpdateArgs, [ ("EIX_LIMIT", "0") ]).Run().Check()

                    ()
                }

            Async.StartAsTask(operation, TaskCreationOptions.None))

    rootCommand.AddArgument remotesArgument

    rootCommand.AddOption gitPullFromRemoteOption
    rootCommand.AddOption gitSignOrigin

    rootCommand.AddOption jobsOption
    rootCommand.AddOption loadAverageOption

    rootCommand.SetHandler rootCommandHandler

    rootCommand.Invoke(argv)
